"""
PySCeS/Core2 - the Next Generation Python Simulator for Cellular Systems
           (http://pysces.sourceforge.net)
Copyright (C) B.G. Olivier
Stellenbosch, 2007-2017.
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa
Author: Brett G. Olivier

PySCeS is Open Source Software distributed under
the GNU GENERAL PUBLIC LICENSE (see docs/GPL)
"""

# Init file used for distutils install