"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

from __future__ import division, print_function
from __future__ import absolute_import
from __future__ import unicode_literals

'''
This file is for "unofficial" user contributed modules. It is imported into PyscesContrib using
 from PyscesContribUser import *
Make sure to use the same module class definitions as used in PyscesContrib i.e. CONTRIB_modulename
to allow for automated module loading and instantiation.
'''

from pysces.version import __version__

__doc__ = '''pysx unofficial user contributed module interface'''
