"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""

squidy =\
"""
     :vzW:.
        :9#@M8v:.    .#MMMMMMM#A;.             .:i,.
     :vv.    tMMMMMM.#MM@YivBMMMMMMMMMMMM@MMMM#Qbz1X
     ..,YEMQ,     7MMMMM#zi        ...  .ii..
  ...      vEMMMb@E:c$MMMMMMMz
   .i;7ZbIY:...UEBMMMMMMMMMMMM#                                ,MMMM.
         .:zZ#MMMZMMMMMMMMMMMMMb   iMMMMMMMMMMMMMMMMMMWt,     YMMMMMM#
      ....       .:c71tt80oM@@MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM@
        .ii7vt0MMMM09ZtnWMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM,
       ...::cnWMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMUMMMM@MMMMMMMc
          .YUc    .;o@M@ZXEM0zB@MM@MMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMMM6
 .YYtn9UIA$@6Z$W@E7:QMM0ZMMMMMMco; 1MMMMMMMMMMMMMMMMMMMMMMMMM#MMMMMMMM#
        ..      .t6MMMMMMMMMMMX            .t0@MMMMM@I,..      iMMMMM$
           ;QMMMMMWS1tQMMMMM@;                                   ibMi
     .ivE@M0v.  it@M@WMMM0c
  .,;Y:    .XE202i. .QM7
      .:Yt2c.      EMA
    ....           MM6  .:i;7z8Q#$EEz11Cv,.
                    #MMMMMMMMMMM@@#MMMMMMMM@@WEU7

Welcome to the Kraken distributed modelling environment
(c) Brett G. Olivier, Stellenbosch 2007-2017.
"""
octopussy =\
"""
                           .Yt90QW$EC.
                        .U@@@W8AzoI0MMW.
                       WMM@WEE88Z980W@MM@
                      MM$Q00bEbbbbEE0W$#MC
                     bMEbbbbbbbbb88bbE00@@
                     M$8bbZ8bb8bbbbbZZb0QM:
                     MBb8ZbbbbbbbbbbEbEEQ@U             ..
                     M@bZ8EbbbbbbbbbbbbEE##           tMMMM@Z;
                     7M8bbbbbbbbbbbbbbbbE@@          6M@#@@MMMMi
                     .Mbb888ZbZZ88Zb8Z88bMZ          iM@00#.
       ;M             @#88bZ86WQEWW8Z88b6M,            @B@M$
       MM.            UM88bb0#tY@,n@QZ8ZWM             ;MBMB
       MM@. .,i:      .ME8b8Q$It@28B8886Mi       .     EBUZ@7
i      MM@MMMMMM.      MQb8Z8600b08ZbbA@1       tMM@QEM@90#M@
@     MMMM@MM#$M0      z@98b8888888bE9WW        @M@@MM@$MMMM.
MS    Y2,.,,,@QBMMZ0MM#@@A8bEEbZQE6ZZb@#@M     .MWQ$W,...Y2.
#MI      ;St YM@@MM@#$00WZ8bbb8E@@69Z#E0#M@   .##bZM.         2A,   .0
E#MM;  6M@B$M    XM$#BE60bZbEb98QWZ9B06Q@@MMM#M#E6IM. .8MM#..MMMMQ;CQ#
:M#@MMMM$@#Q#B   7@Q0W#@0ZZZ9ZZZA9bbM0E$EEQ#MM#B@M0MBWMMB$MMMMBB#MMM@b
.MMMMM@#0.;MW@M#@$9E0Z9Z660W9b889666WB089bQ#WB@@#M@#@#$#@@#$QEE@MMM8,.
.@M6 .MMz  ,MM#BQ#@ZE@@BW##E88bb0$bUAU69b0MB622Q#7;QQb@MM1S@@@#7Q#.
       .    MM#@MI     iY0WAb#Q8EW##QE0EZ87i80EA    .   .   ..
                  Y21c. .EE8W.0#6AU#0. ;8b8ZQ#2.:U#@@@S
        c@@@c    $M@@@MM@09@i .Mb9EE@@$$QbEb690@M@W@M@MM
       .MM@@MM06#MQ$@#WQ0EZM.  8MM  .M@00#EMM@#B$@@MMMBMA
    :tAWME0#@@@@B8@MMMBIW@@B   MMQ    QQQ0 cMMUZS. :i @Mo
  .MMMMWZ6WZ.#M@W@W:E;                               :#i
  7MBW@@$@@  ZMZt9
Welcome to the Kraken distributed modelling environment
(c) Brett G. Olivier, Stellenbosch 2007-2017.
"""