"""
PySCeS - Python Simulator for Cellular Systems (http://pysces.sourceforge.net)

Copyright (C) 2004-2022 B.G. Olivier, J.M. Rohwer, J.-H.S Hofmeyr all rights reserved,

Brett G. Olivier (bgoli@users.sourceforge.net)
Triple-J Group for Molecular Cell Physiology
Stellenbosch University, South Africa.

Permission to use, modify, and distribute this software is given under the
terms of the PySceS (BSD style) license. See LICENSE.txt that came with
this distribution for specifics.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
Brett G. Olivier
"""
"""
METATOOL is a C program developed from 1998 to 2000 by Thomas Pfeiffer
(Berlin) in cooperation with Stefan Schuster and Ferdinand
Moldenhauer (Berlin) and Juan Carlos Nuno (Madrid).

See also
http://www.biologie.hu-berlin.de/biophysics/Theory/tpfeiffer/metatool.html.
"""

# Init file used for distutils install
